package javax.swing.text;

import javax.swing.undo.AbstractUndoableEdit;
import javax.swing.UIManager;
import javax.swing.undo.CannotUndoException;
import javax.swing.undo.CannotRedoException;
import javax.swing.undo.CompoundEdit;
import java.io.ObjectOutputStream;
import java.util.Enumeration;
import java.io.UnsupportedEncodingException;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import javax.swing.tree.TreeNode;
import java.io.IOException;
import java.io.ObjectInputValidation;
import java.io.ObjectInputStream;
import java.io.PrintStream;
import sun.font.BidiUtils;
import java.text.Bidi;
import java.util.Vector;
import sun.swing.SwingUtilities2;
import javax.swing.undo.UndoableEdit;
import java.awt.font.TextAttribute;
import java.util.EventListener;
import javax.swing.event.UndoableEditListener;
import javax.swing.event.UndoableEditEvent;
import javax.swing.event.DocumentListener;
import javax.swing.event.DocumentEvent;
import java.util.Hashtable;
import java.security.AccessController;
import java.security.PrivilegedAction;
import javax.swing.event.EventListenerList;
import java.util.Dictionary;
import java.io.Serializable;

public abstract class AbstractDocument implements Document, Serializable
{
    private transient int numReaders;
    private transient Thread currWriter;
    private transient int numWriters;
    private transient boolean notifyingListeners;
    private static Boolean defaultI18NProperty;
    private Dictionary<Object, Object> documentProperties;
    protected EventListenerList listenerList;
    private AbstractDocument.Content data;
    private AbstractDocument.AttributeContext context;
    private transient AbstractDocument.BranchElement bidiRoot;
    private DocumentFilter documentFilter;
    private transient DocumentFilter.FilterBypass filterBypass;
    private static final String BAD_LOCK_STATE = "document lock failure";
    protected static final String BAD_LOCATION = "document location failure";
    public static final String ParagraphElementName = "paragraph";
    public static final String ContentElementName = "content";
    public static final String SectionElementName = "section";
    public static final String BidiElementName = "bidi level";
    public static final String ElementNameAttribute = "$ename";
    static final String I18NProperty = "i18n";
    static final Object MultiByteProperty;
    static final String AsyncLoadPriority = "load priority";
    
    protected AbstractDocument(AbstractDocument.Content content) {
        this(content, StyleContext.getDefaultStyleContext());
    }
    
    protected AbstractDocument(AbstractDocument.Content data, AbstractDocument.AttributeContext context) {
        this.documentProperties = null;
        this.listenerList = new EventListenerList();
        this.data = data;
        this.context = context;
        this.bidiRoot = new AbstractDocument.BidiRootElement();
        if (AbstractDocument.defaultI18NProperty == null) {
            String s = (String)AccessController.doPrivileged(new PrivilegedAction<String>() {
                @Override
                public String run() {
                    return System.getProperty("i18n");
                }
            });
            if (s != null) {
                AbstractDocument.defaultI18NProperty = Boolean.valueOf(s);
            }
            else {
                AbstractDocument.defaultI18NProperty = Boolean.FALSE;
            }
        }
        this.putProperty("i18n", AbstractDocument.defaultI18NProperty);
        this.writeLock();
        try {
            this.bidiRoot.replace(0, 0, new Element[] { new AbstractDocument.BidiElement(this.bidiRoot, 0, 1, 0) });
        }
        finally {
            this.writeUnlock();
        }
    }
    
    public Dictionary<Object, Object> getDocumentProperties() {
        if (this.documentProperties == null) {
            this.documentProperties = new Hashtable(2);
        }
        return this.documentProperties;
    }
    
    public void setDocumentProperties(Dictionary<Object, Object> documentProperties) {
        this.documentProperties = documentProperties;
    }
    
    protected void fireInsertUpdate(DocumentEvent documentEvent) {
        this.notifyingListeners = true;
        try {
            Object[] listenerList = this.listenerList.getListenerList();
            for (int i = listenerList.length - 2; i >= 0; i -= 2) {
                if (listenerList[i] == DocumentListener.class) {
                    ((DocumentListener)listenerList[i + 1]).insertUpdate(documentEvent);
                }
            }
        }
        finally {
            this.notifyingListeners = false;
        }
    }
    
    protected void fireChangedUpdate(DocumentEvent documentEvent) {
        this.notifyingListeners = true;
        try {
            Object[] listenerList = this.listenerList.getListenerList();
            for (int i = listenerList.length - 2; i >= 0; i -= 2) {
                if (listenerList[i] == DocumentListener.class) {
                    ((DocumentListener)listenerList[i + 1]).changedUpdate(documentEvent);
                }
            }
        }
        finally {
            this.notifyingListeners = false;
        }
    }
    
    protected void fireRemoveUpdate(DocumentEvent documentEvent) {
        this.notifyingListeners = true;
        try {
            Object[] listenerList = this.listenerList.getListenerList();
            for (int i = listenerList.length - 2; i >= 0; i -= 2) {
                if (listenerList[i] == DocumentListener.class) {
                    ((DocumentListener)listenerList[i + 1]).removeUpdate(documentEvent);
                }
            }
        }
        finally {
            this.notifyingListeners = false;
        }
    }
    
    protected void fireUndoableEditUpdate(UndoableEditEvent undoableEditEvent) {
        Object[] listenerList = this.listenerList.getListenerList();
        for (int i = listenerList.length - 2; i >= 0; i -= 2) {
            if (listenerList[i] == UndoableEditListener.class) {
                ((UndoableEditListener)listenerList[i + 1]).undoableEditHappened(undoableEditEvent);
            }
        }
    }
    
    public <T extends EventListener> T[] getListeners(Class<T> t) {
        return this.listenerList.getListeners(t);
    }
    
    public int getAsynchronousLoadPriority() {
        Integer n = (Integer)this.getProperty("load priority");
        if (n != null) {
            return (int)n;
        }
        return -1;
    }
    
    public void setAsynchronousLoadPriority(int i) {
        this.putProperty("load priority", (i >= 0) ? Integer.valueOf(i) : null);
    }
    
    public void setDocumentFilter(DocumentFilter documentFilter) {
        this.documentFilter = documentFilter;
    }
    
    public DocumentFilter getDocumentFilter() {
        return this.documentFilter;
    }
    
    @Override
    public void render(Runnable runnable) {
        this.readLock();
        try {
            runnable.run();
        }
        finally {
            this.readUnlock();
        }
    }
    
    @Override
    public int getLength() {
        return this.data.length() - 1;
    }
    
    @Override
    public void addDocumentListener(DocumentListener l) {
        this.listenerList.add(DocumentListener.class, l);
    }
    
    @Override
    public void removeDocumentListener(DocumentListener l) {
        this.listenerList.remove(DocumentListener.class, l);
    }
    
    public DocumentListener[] getDocumentListeners() {
        return (DocumentListener[])this.listenerList.getListeners(DocumentListener.class);
    }
    
    @Override
    public void addUndoableEditListener(UndoableEditListener l) {
        this.listenerList.add(UndoableEditListener.class, l);
    }
    
    @Override
    public void removeUndoableEditListener(UndoableEditListener l) {
        this.listenerList.remove(UndoableEditListener.class, l);
    }
    
    public UndoableEditListener[] getUndoableEditListeners() {
        return (UndoableEditListener[])this.listenerList.getListeners(UndoableEditListener.class);
    }
    
    @Override
    public final Object getProperty(Object o) {
        return this.getDocumentProperties().get(o);
    }
    
    @Override
    public final void putProperty(Object o, Object o2) {
        if (o2 != null) {
            this.getDocumentProperties().put(o, o2);
        }
        else {
            this.getDocumentProperties().remove(o);
        }
        if (o == TextAttribute.RUN_DIRECTION && Boolean.TRUE.equals(this.getProperty("i18n"))) {
            this.writeLock();
            try {
                this.updateBidi(new AbstractDocument.DefaultDocumentEvent(0, this.getLength(), DocumentEvent.EventType.INSERT));
            }
            finally {
                this.writeUnlock();
            }
        }
    }
    
    @Override
    public void remove(int n, int n2) throws BadLocationException {
        DocumentFilter documentFilter = this.getDocumentFilter();
        this.writeLock();
        try {
            if (documentFilter != null) {
                documentFilter.remove(this.getFilterBypass(), n, n2);
            }
            else {
                this.handleRemove(n, n2);
            }
        }
        finally {
            this.writeUnlock();
        }
    }
    
    void handleRemove(int n, int n2) throws BadLocationException {
        if (n2 > 0) {
            if (n < 0 || n + n2 > this.getLength()) {
                throw new BadLocationException("Invalid remove", this.getLength() + 1);
            }
            AbstractDocument.DefaultDocumentEvent edit = new AbstractDocument.DefaultDocumentEvent(n, n2, DocumentEvent.EventType.REMOVE);
            int composedTextElement = Utilities.isComposedTextElement(this, n);
            this.removeUpdate(edit);
            UndoableEdit remove = this.data.remove(n, n2);
            if (remove != null) {
                edit.addEdit(remove);
            }
            this.postRemoveUpdate(edit);
            edit.end();
            this.fireRemoveUpdate(edit);
            if (remove != null && composedTextElement == 0) {
                this.fireUndoableEditUpdate(new UndoableEditEvent(this, edit));
            }
        }
    }
    
    public void replace(int n, int n2, String s, AttributeSet set) throws BadLocationException {
        if (n2 == 0 && (s == null || s.length() == 0)) {
            return;
        }
        DocumentFilter documentFilter = this.getDocumentFilter();
        this.writeLock();
        try {
            if (documentFilter != null) {
                documentFilter.replace(this.getFilterBypass(), n, n2, s, set);
            }
            else {
                if (n2 > 0) {
                    this.remove(n, n2);
                }
                if (s != null && s.length() > 0) {
                    this.insertString(n, s, set);
                }
            }
        }
        finally {
            this.writeUnlock();
        }
        return;
    }
    
    @Override
    public void insertString(int n, String s, AttributeSet set) throws BadLocationException {
        if (s == null || s.length() == 0) {
            return;
        }
        DocumentFilter documentFilter = this.getDocumentFilter();
        this.writeLock();
        try {
            if (documentFilter != null) {
                documentFilter.insertString(this.getFilterBypass(), n, s, set);
            }
            else {
                this.handleInsertString(n, s, set);
            }
        }
        finally {
            this.writeUnlock();
        }
        return;
    }
    
    private void handleInsertString(int n, String s, AttributeSet set) throws BadLocationException {
        if (s == null || s.length() == 0) {
            return;
        }
        UndoableEdit insertString = this.data.insertString(n, s);
        AbstractDocument.DefaultDocumentEvent edit = new AbstractDocument.DefaultDocumentEvent(n, s.length(), DocumentEvent.EventType.INSERT);
        if (insertString != null) {
            edit.addEdit(insertString);
        }
        if (this.getProperty("i18n").equals(Boolean.FALSE)) {
            Object property = this.getProperty(TextAttribute.RUN_DIRECTION);
            if (property != null && property.equals(TextAttribute.RUN_DIRECTION_RTL)) {
                this.putProperty("i18n", Boolean.TRUE);
            }
            else {
                char[] charArray = s.toCharArray();
                if (SwingUtilities2.isComplexLayout(charArray, 0, charArray.length)) {
                    this.putProperty("i18n", Boolean.TRUE);
                }
            }
        }
        this.insertUpdate(edit, set);
        edit.end();
        this.fireInsertUpdate(edit);
        if (insertString != null && (set == null || !set.isDefined(StyleConstants.ComposedTextAttribute))) {
            this.fireUndoableEditUpdate(new UndoableEditEvent(this, edit));
        }
        return;
    }
    
    @Override
    public String getText(int n, int n2) throws BadLocationException {
        if (n2 < 0) {
            throw new BadLocationException("Length must be positive", n2);
        }
        return this.data.getString(n, n2);
    }
    
    @Override
    public void getText(int n, int n2, Segment segment) throws BadLocationException {
        if (n2 < 0) {
            throw new BadLocationException("Length must be positive", n2);
        }
        this.data.getChars(n, n2, segment);
        return;
    }
    
    @Override
    public synchronized Position createPosition(int n) throws BadLocationException {
        return this.data.createPosition(n);
    }
    
    @Override
    public final Position getStartPosition() {
        Position position;
        try {
            position = this.createPosition(0);
        }
        catch (BadLocationException ex) {
            position = null;
        }
        return position;
    }
    
    @Override
    public final Position getEndPosition() {
        Position position;
        try {
            position = this.createPosition(this.data.length());
        }
        catch (BadLocationException ex) {
            position = null;
        }
        return position;
    }
    
    @Override
    public Element[] getRootElements() {
        return new Element[] { this.getDefaultRootElement(), this.getBidiRootElement() };
    }
    
    @Override
    public abstract Element getDefaultRootElement();
    
    private DocumentFilter.FilterBypass getFilterBypass() {
        if (this.filterBypass == null) {
            this.filterBypass = new AbstractDocument.DefaultFilterBypass();
        }
        return this.filterBypass;
    }
    
    public Element getBidiRootElement() {
        return this.bidiRoot;
    }
    
    static boolean isLeftToRight(Document document, int n, int n2) {
        if (Boolean.TRUE.equals(document.getProperty("i18n")) && document instanceof AbstractDocument) {
            Element bidiRootElement = ((AbstractDocument)document).getBidiRootElement();
            Element element = bidiRootElement.getElement(bidiRootElement.getElementIndex(n));
            if (element.getEndOffset() >= n2) {
                return StyleConstants.getBidiLevel(element.getAttributes()) % 2 == 0;
            }
        }
        return true;
    }
    
    public abstract Element getParagraphElement(int p0);
    
    protected final AbstractDocument.AttributeContext getAttributeContext() {
        return this.context;
    }
    
    protected void insertUpdate(AbstractDocument.DefaultDocumentEvent defaultDocumentEvent, AttributeSet set) {
        if (this.getProperty("i18n").equals(Boolean.TRUE)) {
            this.updateBidi(defaultDocumentEvent);
        }
        if (defaultDocumentEvent.type == DocumentEvent.EventType.INSERT && defaultDocumentEvent.getLength() > 0 && !Boolean.TRUE.equals(this.getProperty(AbstractDocument.MultiByteProperty))) {
            Segment sharedSegment = SegmentCache.getSharedSegment();
            Label_0112: {
                try {
                    this.getText(defaultDocumentEvent.getOffset(), defaultDocumentEvent.getLength(), sharedSegment);
                    sharedSegment.first();
                    while (sharedSegment.current() <= 'ÿ') {
                        if (sharedSegment.next() != '\uffff') {
                            continue;
                        }
                        break Label_0112;
                    }
                    this.putProperty(AbstractDocument.MultiByteProperty, Boolean.TRUE);
                }
                catch (BadLocationException ex) {}
            }
            SegmentCache.releaseSharedSegment(sharedSegment);
        }
    }
    
    protected void removeUpdate(AbstractDocument.DefaultDocumentEvent defaultDocumentEvent) {
    }
    
    protected void postRemoveUpdate(AbstractDocument.DefaultDocumentEvent defaultDocumentEvent) {
        if (this.getProperty("i18n").equals(Boolean.TRUE)) {
            this.updateBidi(defaultDocumentEvent);
        }
    }
    
    void updateBidi(AbstractDocument.DefaultDocumentEvent defaultDocumentEvent) {
        int n;
        int n2;
        if (defaultDocumentEvent.type == DocumentEvent.EventType.INSERT || defaultDocumentEvent.type == DocumentEvent.EventType.CHANGE) {
            int offset = defaultDocumentEvent.getOffset();
            int n3 = offset + defaultDocumentEvent.getLength();
            n = this.getParagraphElement(offset).getStartOffset();
            n2 = this.getParagraphElement(n3).getEndOffset();
        }
        else if (defaultDocumentEvent.type == DocumentEvent.EventType.REMOVE) {
            Element paragraphElement = this.getParagraphElement(defaultDocumentEvent.getOffset());
            n = paragraphElement.getStartOffset();
            n2 = paragraphElement.getEndOffset();
        }
        else {
            throw new Error("Internal error: unknown event type.");
        }
        byte[] calculateBidiLevels = this.calculateBidiLevels(n, n2);
        Vector o = new Vector();
        int elementIndex = 0;
        if (n > 0) {
            Element element = this.bidiRoot.getElement(elementIndex = this.bidiRoot.getElementIndex(n - 1));
            int bidiLevel = StyleConstants.getBidiLevel(element.getAttributes());
            if (bidiLevel == calculateBidiLevels[0]) {
                n = element.getStartOffset();
            }
            else if (element.getEndOffset() > n) {
                o.addElement(new AbstractDocument.BidiElement(this.bidiRoot, element.getStartOffset(), n, bidiLevel));
            }
            else {
                ++elementIndex;
            }
        }
        int i;
        for (i = 0; i < calculateBidiLevels.length && calculateBidiLevels[i] == calculateBidiLevels[0]; ++i) {}
        Object obj = null;
        int elementIndex2 = this.bidiRoot.getElementCount() - 1;
        if (n2 <= this.getLength()) {
            Element element2 = this.bidiRoot.getElement(elementIndex2 = this.bidiRoot.getElementIndex(n2));
            int bidiLevel2 = StyleConstants.getBidiLevel(element2.getAttributes());
            if (bidiLevel2 == calculateBidiLevels[calculateBidiLevels.length - 1]) {
                n2 = element2.getEndOffset();
            }
            else if (element2.getStartOffset() < n2) {
                obj = new AbstractDocument.BidiElement(this.bidiRoot, n2, element2.getEndOffset(), bidiLevel2);
            }
            else {
                --elementIndex2;
            }
        }
        int length;
        for (length = calculateBidiLevels.length; length > i && calculateBidiLevels[length - 1] == calculateBidiLevels[calculateBidiLevels.length - 1]; --length) {}
        if (i == length && calculateBidiLevels[0] == calculateBidiLevels[calculateBidiLevels.length - 1]) {
            o.addElement(new AbstractDocument.BidiElement(this.bidiRoot, n, n2, (int)calculateBidiLevels[0]));
        }
        else {
            o.addElement(new AbstractDocument.BidiElement(this.bidiRoot, n, i + n, (int)calculateBidiLevels[0]));
            int n4;
            for (; i < length; i = n4) {
                for (n4 = i; n4 < calculateBidiLevels.length && calculateBidiLevels[n4] == calculateBidiLevels[i]; ++n4) {}
                o.addElement(new AbstractDocument.BidiElement(this.bidiRoot, n + i, n + n4, (int)calculateBidiLevels[i]));
            }
            o.addElement(new AbstractDocument.BidiElement(this.bidiRoot, length + n, n2, (int)calculateBidiLevels[calculateBidiLevels.length - 1]));
        }
        if (obj != null) {
            o.addElement(obj);
        }
        int n5 = 0;
        if (this.bidiRoot.getElementCount() > 0) {
            n5 = elementIndex2 - elementIndex + 1;
        }
        Element[] array = new Element[n5];
        for (int j = 0; j < n5; ++j) {
            array[j] = this.bidiRoot.getElement(elementIndex + j);
        }
        Element[] anArray = new Element[o.size()];
        o.copyInto(anArray);
        defaultDocumentEvent.addEdit(new AbstractDocument.ElementEdit(this.bidiRoot, elementIndex, array, anArray));
        this.bidiRoot.replace(elementIndex, array.length, anArray);
    }
    
    private byte[] calculateBidiLevels(int i, int n) {
        byte[] levels = new byte[n - i];
        int start = 0;
        Boolean obj = null;
        Object property = this.getProperty(TextAttribute.RUN_DIRECTION);
        if ((property instanceof Boolean)) {
            obj = (Boolean)property;
        }
        while (i < n) {
            Element paragraphElement = this.getParagraphElement(i);
            int startOffset = paragraphElement.getStartOffset();
            int endOffset = paragraphElement.getEndOffset();
            Object attribute = paragraphElement.getAttributes().getAttribute(TextAttribute.RUN_DIRECTION);
            if ((attribute instanceof Boolean)) {
                obj = (Boolean)attribute;
            }
            Segment sharedSegment = SegmentCache.getSharedSegment();
            try {
                this.getText(startOffset, endOffset - startOffset, sharedSegment);
            }
            catch (BadLocationException ex) {
                throw new Error("Internal error: " + ex.toString());
            }
            int flags = -2;
            if (obj != null) {
                if (TextAttribute.RUN_DIRECTION_LTR.equals(obj)) {
                    flags = 0;
                }
                else {
                    flags = 1;
                }
            }
            Bidi bidi = new Bidi(sharedSegment.array, sharedSegment.offset, null, 0, sharedSegment.count, flags);
            BidiUtils.getLevels(bidi, levels, start);
            start += bidi.getLength();
            i = paragraphElement.getEndOffset();
            SegmentCache.releaseSharedSegment(sharedSegment);
        }
        if (start != levels.length) {
            throw new Error("levelsEnd assertion failed.");
        }
        return levels;
    }
    
    public void dump(PrintStream printStream) {
        Element defaultRootElement = this.getDefaultRootElement();
        if ((defaultRootElement instanceof AbstractDocument.AbstractElement)) {
            ((AbstractDocument.AbstractElement)defaultRootElement).dump(printStream, 0);
        }
        this.bidiRoot.dump(printStream, 0);
    }
    
    protected final AbstractDocument.Content getContent() {
        return this.data;
    }
    
    protected Element createLeafElement(Element element, AttributeSet set, int n, int n2) {
        return new AbstractDocument.LeafElement(element, set, n, n2);
    }
    
    protected Element createBranchElement(Element element, AttributeSet set) {
        return new AbstractDocument.BranchElement(element, set);
    }
    
    protected final synchronized Thread getCurrentWriter() {
        return this.currWriter;
    }
    
    protected final synchronized void writeLock() {
        try {
            while (this.numReaders > 0 || this.currWriter != null) {
                if (Thread.currentThread() == this.currWriter) {
                    if (this.notifyingListeners) {
                        throw new IllegalStateException("Attempt to mutate in notification");
                    }
                    ++this.numWriters;
                    return;
                }
                else {
                    this.wait();
                }
            }
            this.currWriter = Thread.currentThread();
            this.numWriters = 1;
        }
        catch (InterruptedException ex) {
            throw new Error("Interrupted attempt to acquire write lock");
        }
    }
    
    protected final synchronized void writeUnlock() {
        int numWriters = this.numWriters - 1;
        this.numWriters = numWriters;
        if (numWriters <= 0) {
            this.numWriters = 0;
            this.currWriter = null;
            this.notifyAll();
        }
    }
    
    public final synchronized void readLock() {
        try {
            while (this.currWriter != null) {
                if (this.currWriter == Thread.currentThread()) {
                    return;
                }
                this.wait();
            }
            ++this.numReaders;
        }
        catch (InterruptedException ex) {
            throw new Error("Interrupted attempt to acquire read lock");
        }
    }
    
    public final synchronized void readUnlock() {
        if (this.currWriter == Thread.currentThread()) {
            return;
        }
        if (this.numReaders <= 0) {
            throw new StateInvariantError("document lock failure");
        }
        --this.numReaders;
        this.notify();
        return;
    }
    
    private void readObject(ObjectInputStream objectInputStream) throws ClassNotFoundException, IOException {
        objectInputStream.defaultReadObject();
        this.listenerList = new EventListenerList();
        this.bidiRoot = new AbstractDocument.BidiRootElement();
        try {
            this.writeLock();
            this.bidiRoot.replace(0, 0, new Element[] { new AbstractDocument.BidiElement(this.bidiRoot, 0, 1, 0) });
        }
        finally {
            this.writeUnlock();
        }
        objectInputStream.registerValidation(new ObjectInputValidation() {
            @Override
            public void validateObject() {
                try {
                    AbstractDocument.this.writeLock();
                    AbstractDocument.this.updateBidi(new AbstractDocument.DefaultDocumentEvent(0, AbstractDocument.this.getLength(), DocumentEvent.EventType.INSERT));
                }
                finally {
                    AbstractDocument.this.writeUnlock();
                }
            }
        }, 0);
    }
    
    static void access$200(AbstractDocument abstractDocument, int n, String s, AttributeSet set) throws BadLocationException {
        abstractDocument.handleInsertString(n, s, set);
    }
    
    static {
        MultiByteProperty = "multiByte";
    }
    
    public abstract class AbstractElement implements Element, MutableAttributeSet, Serializable, TreeNode
    {
        private Element parent;
        private transient AttributeSet attributes;
        
        public AbstractElement(Element parent, AttributeSet set) {
            this.parent = parent;
            this.attributes = AbstractDocument.this.getAttributeContext().getEmptySet();
            if (set != null) {
                this.addAttributes(set);
            }
        }
        
        private final void indent(PrintWriter printWriter, int n) {
            for (int i = 0; i < n; ++i) {
                printWriter.print("  ");
            }
        }
        
        public void dump(PrintStream printStream, int n) {
            PrintWriter printWriter;
            try {
                printWriter = new PrintWriter(new OutputStreamWriter(printStream, "JavaEsc"), true);
            }
            catch (UnsupportedEncodingException ex) {
                printWriter = new PrintWriter(printStream, true);
            }
            this.indent(printWriter, n);
            if (this.getName() == null) {
                printWriter.print("<??");
            }
            else {
                printWriter.print("<" + this.getName());
            }
            if (this.getAttributeCount() > 0) {
                printWriter.println("");
                Enumeration<Object> attributeNames = this.attributes.getAttributeNames();
                while (attributeNames.hasMoreElements()) {
                    Object nextElement = attributeNames.nextElement();
                    this.indent(printWriter, n + 1);
                    printWriter.println(nextElement + "=" + this.getAttribute(nextElement));
                }
                this.indent(printWriter, n);
            }
            printWriter.println(">");
            if (this.isLeaf()) {
                this.indent(printWriter, n + 1);
                printWriter.print("[" + this.getStartOffset() + "," + this.getEndOffset() + "]");
                AbstractDocument.Content content = AbstractDocument.this.getContent();
                try {
                    String str = content.getString(this.getStartOffset(), this.getEndOffset() - this.getStartOffset());
                    if (str.length() > 40) {
                        str = str.substring(0, 40) + "...";
                    }
                    printWriter.println("[" + str + "]");
                }
                catch (BadLocationException ex2) {}
            }
            else {
                for (int elementCount = this.getElementCount(), i = 0; i < elementCount; ++i) {
                    ((AbstractDocument.AbstractElement)this.getElement(i)).dump(printStream, n + 1);
                }
            }
        }
        
        @Override
        public int getAttributeCount() {
            return this.attributes.getAttributeCount();
        }
        
        @Override
        public boolean isDefined(Object o) {
            return this.attributes.isDefined(o);
        }
        
        @Override
        public boolean isEqual(AttributeSet set) {
            return this.attributes.isEqual(set);
        }
        
        @Override
        public AttributeSet copyAttributes() {
            return this.attributes.copyAttributes();
        }
        
        @Override
        public Object getAttribute(Object o) {
            Object o2 = this.attributes.getAttribute(o);
            if (o2 == null) {
                AttributeSet set = (this.parent != null) ? this.parent.getAttributes() : null;
                if (set != null) {
                    o2 = set.getAttribute(o);
                }
            }
            return o2;
        }
        
        @Override
        public Enumeration<?> getAttributeNames() {
            return this.attributes.getAttributeNames();
        }
        
        @Override
        public boolean containsAttribute(Object o, Object o2) {
            return this.attributes.containsAttribute(o, o2);
        }
        
        @Override
        public boolean containsAttributes(AttributeSet set) {
            return this.attributes.containsAttributes(set);
        }
        
        @Override
        public AttributeSet getResolveParent() {
            AttributeSet set = this.attributes.getResolveParent();
            if (set == null && this.parent != null) {
                set = this.parent.getAttributes();
                return set;
            }
            return set;
        }
        
        @Override
        public void addAttribute(Object o, Object o2) {
            this.checkForIllegalCast();
            this.attributes = AbstractDocument.this.getAttributeContext().addAttribute(this.attributes, o, o2);
        }
        
        @Override
        public void addAttributes(AttributeSet set) {
            this.checkForIllegalCast();
            this.attributes = AbstractDocument.this.getAttributeContext().addAttributes(this.attributes, set);
        }
        
        @Override
        public void removeAttribute(Object o) {
            this.checkForIllegalCast();
            this.attributes = AbstractDocument.this.getAttributeContext().removeAttribute(this.attributes, o);
        }
        
        @Override
        public void removeAttributes(Enumeration<?> enumeration) {
            this.checkForIllegalCast();
            this.attributes = AbstractDocument.this.getAttributeContext().removeAttributes(this.attributes, enumeration);
        }
        
        @Override
        public void removeAttributes(AttributeSet set) {
            this.checkForIllegalCast();
            AbstractDocument.AttributeContext attributeContext = AbstractDocument.this.getAttributeContext();
            if (set == this) {
                this.attributes = attributeContext.getEmptySet();
            }
            else {
                this.attributes = attributeContext.removeAttributes(this.attributes, set);
            }
        }
        
        @Override
        public void setResolveParent(AttributeSet set) {
            this.checkForIllegalCast();
            AbstractDocument.AttributeContext attributeContext = AbstractDocument.this.getAttributeContext();
            if (set != null) {
                this.attributes = attributeContext.addAttribute(this.attributes, StyleConstants.ResolveAttribute, set);
            }
            else {
                this.attributes = attributeContext.removeAttribute(this.attributes, StyleConstants.ResolveAttribute);
            }
        }
        
        private final void checkForIllegalCast() {
            Thread currentWriter = AbstractDocument.this.getCurrentWriter();
            if (currentWriter == null || currentWriter != Thread.currentThread()) {
                throw new StateInvariantError("Illegal cast to MutableAttributeSet");
            }
            return;
        }
        
        @Override
        public Document getDocument() {
            return AbstractDocument.this;
        }
        
        @Override
        public Element getParentElement() {
            return this.parent;
        }
        
        @Override
        public AttributeSet getAttributes() {
            return this;
        }
        
        @Override
        public String getName() {
            if (this.attributes.isDefined("$ename")) {
                return (String)this.attributes.getAttribute("$ename");
            }
            return null;
        }
        
        @Override
        public abstract int getStartOffset();
        
        @Override
        public abstract int getEndOffset();
        
        @Override
        public abstract Element getElement(int p0);
        
        @Override
        public abstract int getElementCount();
        
        @Override
        public abstract int getElementIndex(int p0);
        
        @Override
        public abstract boolean isLeaf();
        
        @Override
        public TreeNode getChildAt(int n) {
            return (TreeNode)this.getElement(n);
        }
        
        @Override
        public int getChildCount() {
            return this.getElementCount();
        }
        
        @Override
        public TreeNode getParent() {
            return (TreeNode)this.getParentElement();
        }
        
        @Override
        public int getIndex(TreeNode treeNode) {
            int i = this.getChildCount() - 1;
            while (i >= 0) {
                if (this.getChildAt(i) == treeNode) {
                    return i;
                }
                --i;
            }
            return -1;
        }
        
        @Override
        public abstract boolean getAllowsChildren();
        
        @Override
        public abstract Enumeration children();
        
        private void writeObject(ObjectOutputStream objectOutputStream) throws IOException {
            objectOutputStream.defaultWriteObject();
            StyleContext.writeAttributeSet(objectOutputStream, this.attributes);
        }
        
        private void readObject(ObjectInputStream objectInputStream) throws ClassNotFoundException, IOException {
            objectInputStream.defaultReadObject();
            SimpleAttributeSet set = new SimpleAttributeSet();
            StyleContext.readAttributeSet(objectInputStream, set);
            this.attributes = AbstractDocument.this.getAttributeContext().addAttributes(SimpleAttributeSet.EMPTY, set);
        }
    }
    
    public class BranchElement extends AbstractDocument.AbstractElement
    {
        private AbstractDocument.AbstractElement[] children;
        private int nchildren;
        private int lastIndex;
        
        public BranchElement(Element element, AttributeSet set) {
            super(element, set);
            this.children = new AbstractDocument.AbstractElement[1];
            this.nchildren = 0;
            this.lastIndex = -1;
        }
        
        public Element positionToElement(int n) {
            AbstractDocument.AbstractElement abstractElement = this.children[this.getElementIndex(n)];
            int startOffset = abstractElement.getStartOffset();
            int endOffset = abstractElement.getEndOffset();
            if (n >= startOffset && n < endOffset) {
                return abstractElement;
            }
            return null;
        }
        
        public void replace(int n, int n2, Element[] array) {
            int n3 = array.length - n2;
            int n4 = n + n2;
            int n5 = this.nchildren - n4;
            int n6 = n4 + n3;
            if (this.nchildren + n3 >= this.children.length) {
                AbstractDocument.AbstractElement[] children = new AbstractDocument.AbstractElement[Math.max(2 * this.children.length, this.nchildren + n3)];
                System.arraycopy(this.children, 0, children, 0, n);
                System.arraycopy(array, 0, children, n, array.length);
                System.arraycopy(this.children, n4, children, n6, n5);
                this.children = children;
            }
            else {
                System.arraycopy(this.children, n4, this.children, n6, n5);
                System.arraycopy(array, 0, this.children, n, array.length);
            }
            this.nchildren += n3;
        }
        
        @Override
        public String toString() {
            return "BranchElement(" + this.getName() + ") " + this.getStartOffset() + "," + this.getEndOffset() + "\n";
        }
        
        @Override
        public String getName() {
            String name = super.getName();
            if (name == null) {
                name = "paragraph";
                return name;
            }
            return name;
        }
        
        @Override
        public int getStartOffset() {
            return this.children[0].getStartOffset();
        }
        
        @Override
        public int getEndOffset() {
            return ((this.nchildren > 0) ? this.children[this.nchildren - 1] : this.children[0]).getEndOffset();
        }
        
        @Override
        public Element getElement(int n) {
            if (n < this.nchildren) {
                return this.children[n];
            }
            return null;
        }
        
        @Override
        public int getElementCount() {
            return this.nchildren;
        }
        
        @Override
        public int getElementIndex(int n) {
            int i = 0;
            int lastIndex = this.nchildren - 1;
            int n2 = 0;
            int n3 = this.getStartOffset();
            if (this.nchildren == 0) {
                return 0;
            }
            if (n >= this.getEndOffset()) {
                return this.nchildren - 1;
            }
            if (this.lastIndex >= i && this.lastIndex <= lastIndex) {
                AbstractDocument.AbstractElement abstractElement = this.children[this.lastIndex];
                n3 = abstractElement.getStartOffset();
                int endOffset = abstractElement.getEndOffset();
                if (n >= n3 && n < endOffset) {
                    return this.lastIndex;
                }
                if (n < n3) {
                    lastIndex = this.lastIndex;
                }
                else {
                    i = this.lastIndex;
                }
            }
            while (i <= lastIndex) {
                n2 = i + (lastIndex - i) / 2;
                AbstractDocument.AbstractElement abstractElement2 = this.children[n2];
                n3 = abstractElement2.getStartOffset();
                int endOffset2 = abstractElement2.getEndOffset();
                if (n >= n3 && n < endOffset2) {
                    return this.lastIndex = n2;
                }
                if (n < n3) {
                    lastIndex = n2 - 1;
                }
                else {
                    i = n2 + 1;
                }
            }
            if (n < n3) {
            }
            else {
                ++n2;
            }
            return this.lastIndex = n2;
        }
        
        @Override
        public boolean isLeaf() {
            return false;
        }
        
        @Override
        public boolean getAllowsChildren() {
            return true;
        }
        
        @Override
        public Enumeration children() {
            if (this.nchildren == 0) {
                return null;
            }
            Vector vector = new Vector(this.nchildren);
            for (int i = 0; i < this.nchildren; ++i) {
                vector.addElement(this.children[i]);
            }
            return vector.elements();
        }
    }
    
    public class LeafElement extends AbstractDocument.AbstractElement
    {
        private transient Position p0;
        private transient Position p1;
        
        public LeafElement(Element element, AttributeSet set, int n, int n2) {
            super(element, set);
            try {
                this.p0 = AbstractDocument.this.createPosition(n);
                this.p1 = AbstractDocument.this.createPosition(n2);
            }
            catch (BadLocationException ex) {
                this.p0 = null;
                this.p1 = null;
                throw new StateInvariantError("Can't create Position references");
            }
        }
        
        @Override
        public String toString() {
            return "LeafElement(" + this.getName() + ") " + this.p0 + "," + this.p1 + "\n";
        }
        
        @Override
        public int getStartOffset() {
            return this.p0.getOffset();
        }
        
        @Override
        public int getEndOffset() {
            return this.p1.getOffset();
        }
        
        @Override
        public String getName() {
            String name = super.getName();
            if (name == null) {
                name = "content";
                return name;
            }
            return name;
        }
        
        @Override
        public int getElementIndex(int n) {
            return -1;
        }
        
        @Override
        public Element getElement(int n) {
            return null;
        }
        
        @Override
        public int getElementCount() {
            return 0;
        }
        
        @Override
        public boolean isLeaf() {
            return true;
        }
        
        @Override
        public boolean getAllowsChildren() {
            return false;
        }
        
        @Override
        public Enumeration children() {
            return null;
        }
        
        private void writeObject(ObjectOutputStream objectOutputStream) throws IOException {
            objectOutputStream.defaultWriteObject();
            objectOutputStream.writeInt(this.p0.getOffset());
            objectOutputStream.writeInt(this.p1.getOffset());
        }
        
        private void readObject(ObjectInputStream objectInputStream) throws ClassNotFoundException, IOException {
            objectInputStream.defaultReadObject();
            int int1 = objectInputStream.readInt();
            int int2 = objectInputStream.readInt();
            try {
                this.p0 = AbstractDocument.this.createPosition(int1);
                this.p1 = AbstractDocument.this.createPosition(int2);
            }
            catch (BadLocationException ex) {
                this.p0 = null;
                this.p1 = null;
                throw new IOException("Can't restore Position references");
            }
        }
    }
    
    class BidiRootElement extends AbstractDocument.BranchElement
    {
        BidiRootElement() {
            super(null, null);
        }
        
        @Override
        public String getName() {
            return "bidi root";
        }
    }
    
    class BidiElement extends AbstractDocument.LeafElement
    {
        BidiElement(Element element, int n, int n2, int i) {
            super(element, new SimpleAttributeSet(), n, n2);
            this.addAttribute(StyleConstants.BidiLevel, Integer.valueOf(i));
        }
        
        @Override
        public String getName() {
            return "bidi level";
        }
        
        int getLevel() {
            Integer n = (Integer)this.getAttribute(StyleConstants.BidiLevel);
            if (n != null) {
                return (int)n;
            }
            return 0;
        }
        
        boolean isLeftToRight() {
            return this.getLevel() % 2 == 0;
        }
    }
    
    public class DefaultDocumentEvent extends CompoundEdit implements DocumentEvent
    {
        private int offset;
        private int length;
        private Hashtable<Element, DocumentEvent.ElementChange> changeLookup;
        private DocumentEvent.EventType type;
        
        public DefaultDocumentEvent(int offset, int length, DocumentEvent.EventType type) {
            this.offset = offset;
            this.length = length;
            this.type = type;
        }
        
        @Override
        public String toString() {
            return this.edits.toString();
        }
        
        @Override
        public boolean addEdit(UndoableEdit anEdit) {
            if (this.changeLookup == null && this.edits.size() > 10) {
                this.changeLookup = new Hashtable();
                for (int size = this.edits.size(), i = 0; i < size; ++i) {
                    Object element = this.edits.elementAt(i);
                    if ((element instanceof DocumentEvent.ElementChange)) {
                        DocumentEvent.ElementChange value = (DocumentEvent.ElementChange)element;
                        this.changeLookup.put(value.getElement(), value);
                    }
                }
            }
            if (this.changeLookup != null && anEdit instanceof DocumentEvent.ElementChange) {
                DocumentEvent.ElementChange value2 = (DocumentEvent.ElementChange)anEdit;
                this.changeLookup.put(value2.getElement(), value2);
            }
            return super.addEdit(anEdit);
        }
        
        @Override
        public void redo() throws CannotRedoException {
            AbstractDocument.this.writeLock();
            try {
                super.redo();
                AbstractDocument.UndoRedoDocumentEvent undoRedoDocumentEvent = new AbstractDocument.UndoRedoDocumentEvent(this, false);
                if (this.type == DocumentEvent.EventType.INSERT) {
                    AbstractDocument.this.fireInsertUpdate(undoRedoDocumentEvent);
                }
                else if (this.type == DocumentEvent.EventType.REMOVE) {
                    AbstractDocument.this.fireRemoveUpdate(undoRedoDocumentEvent);
                }
                else {
                    AbstractDocument.this.fireChangedUpdate(undoRedoDocumentEvent);
                }
            }
            finally {
                AbstractDocument.this.writeUnlock();
            }
        }
        
        @Override
        public void undo() throws CannotUndoException {
            AbstractDocument.this.writeLock();
            try {
                super.undo();
                AbstractDocument.UndoRedoDocumentEvent undoRedoDocumentEvent = new AbstractDocument.UndoRedoDocumentEvent(this, true);
                if (this.type == DocumentEvent.EventType.REMOVE) {
                    AbstractDocument.this.fireInsertUpdate(undoRedoDocumentEvent);
                }
                else if (this.type == DocumentEvent.EventType.INSERT) {
                    AbstractDocument.this.fireRemoveUpdate(undoRedoDocumentEvent);
                }
                else {
                    AbstractDocument.this.fireChangedUpdate(undoRedoDocumentEvent);
                }
            }
            finally {
                AbstractDocument.this.writeUnlock();
            }
        }
        
        @Override
        public boolean isSignificant() {
            return true;
        }
        
        @Override
        public String getPresentationName() {
            DocumentEvent.EventType type = this.getType();
            if (type == DocumentEvent.EventType.INSERT) {
                return UIManager.getString("AbstractDocument.additionText");
            }
            if (type == DocumentEvent.EventType.REMOVE) {
                return UIManager.getString("AbstractDocument.deletionText");
            }
            return UIManager.getString("AbstractDocument.styleChangeText");
        }
        
        @Override
        public String getUndoPresentationName() {
            return UIManager.getString("AbstractDocument.undoText") + " " + this.getPresentationName();
        }
        
        @Override
        public String getRedoPresentationName() {
            return UIManager.getString("AbstractDocument.redoText") + " " + this.getPresentationName();
        }
        
        @Override
        public DocumentEvent.EventType getType() {
            return this.type;
        }
        
        @Override
        public int getOffset() {
            return this.offset;
        }
        
        @Override
        public int getLength() {
            return this.length;
        }
        
        @Override
        public Document getDocument() {
            return AbstractDocument.this;
        }
        
        @Override
        public DocumentEvent.ElementChange getChange(Element key) {
            if (this.changeLookup != null) {
                return (DocumentEvent.ElementChange)this.changeLookup.get(key);
            }
            for (int size = this.edits.size(), i = 0; i < size; ++i) {
                Object element = this.edits.elementAt(i);
                if ((element instanceof DocumentEvent.ElementChange)) {
                    DocumentEvent.ElementChange elementChange = (DocumentEvent.ElementChange)element;
                    if (key.equals(elementChange.getElement())) {
                        return elementChange;
                    }
                }
            }
            return null;
        }
    }
    
    class UndoRedoDocumentEvent implements DocumentEvent
    {
        private AbstractDocument.DefaultDocumentEvent src;
        private DocumentEvent.EventType type;
        
        public UndoRedoDocumentEvent(AbstractDocument.DefaultDocumentEvent src, boolean b) {
            this.src = null;
            this.type = null;
            this.src = src;
            if (b) {
                if (src.getType().equals(DocumentEvent.EventType.INSERT)) {
                    this.type = DocumentEvent.EventType.REMOVE;
                }
                else if (src.getType().equals(DocumentEvent.EventType.REMOVE)) {
                    this.type = DocumentEvent.EventType.INSERT;
                }
                else {
                    this.type = src.getType();
                }
            }
            else {
                this.type = src.getType();
            }
        }
        
        public AbstractDocument.DefaultDocumentEvent getSource() {
            return this.src;
        }
        
        @Override
        public int getOffset() {
            return this.src.getOffset();
        }
        
        @Override
        public int getLength() {
            return this.src.getLength();
        }
        
        @Override
        public Document getDocument() {
            return this.src.getDocument();
        }
        
        @Override
        public DocumentEvent.EventType getType() {
            return this.type;
        }
        
        @Override
        public DocumentEvent.ElementChange getChange(Element element) {
            return this.src.getChange(element);
        }
    }
    
    public static class ElementEdit extends AbstractUndoableEdit implements DocumentEvent.ElementChange
    {
        private Element e;
        private int index;
        private Element[] removed;
        private Element[] added;
        
        public ElementEdit(Element e, int index, Element[] removed, Element[] added) {
            this.e = e;
            this.index = index;
            this.removed = removed;
            this.added = added;
        }
        
        @Override
        public Element getElement() {
            return this.e;
        }
        
        @Override
        public int getIndex() {
            return this.index;
        }
        
        @Override
        public Element[] getChildrenRemoved() {
            return this.removed;
        }
        
        @Override
        public Element[] getChildrenAdded() {
            return this.added;
        }
        
        @Override
        public void redo() throws CannotRedoException {
            super.redo();
            Element[] removed = this.removed;
            this.removed = this.added;
            this.added = removed;
            ((AbstractDocument.BranchElement)this.e).replace(this.index, this.removed.length, this.added);
        }
        
        @Override
        public void undo() throws CannotUndoException {
            super.undo();
            ((AbstractDocument.BranchElement)this.e).replace(this.index, this.added.length, this.removed);
            Element[] removed = this.removed;
            this.removed = this.added;
            this.added = removed;
        }
    }
    
    private class DefaultFilterBypass extends DocumentFilter.FilterBypass
    {
        @Override
        public Document getDocument() {
            return AbstractDocument.this;
        }
        
        @Override
        public void remove(int n, int n2) throws BadLocationException {
            AbstractDocument.this.handleRemove(n, n2);
        }
        
        @Override
        public void insertString(int n, String s, AttributeSet set) throws BadLocationException {
            AbstractDocument.access$200(AbstractDocument.this, n, s, set);
        }
        
        @Override
        public void replace(int n, int n2, String s, AttributeSet set) throws BadLocationException {
            AbstractDocument.this.handleRemove(n, n2);
            AbstractDocument.access$200(AbstractDocument.this, n, s, set);
        }
    }
    
    public interface AttributeContext
    {
        public abstract AttributeSet addAttribute(AttributeSet p0, Object p1, Object p2);
        
        public abstract AttributeSet addAttributes(AttributeSet p0, AttributeSet p1);
        
        public abstract AttributeSet removeAttribute(AttributeSet p0, Object p1);
        
        public abstract AttributeSet removeAttributes(AttributeSet p0, Enumeration<?> p1);
        
        public abstract AttributeSet removeAttributes(AttributeSet p0, AttributeSet p1);
        
        public abstract AttributeSet getEmptySet();
        
        public abstract void reclaim(AttributeSet p0);
    }
    
    public interface Content
    {
        public abstract Position createPosition(int p0) throws BadLocationException;
        
        public abstract int length();
        
        public abstract UndoableEdit insertString(int p0, String p1) throws BadLocationException;
        
        public abstract UndoableEdit remove(int p0, int p1) throws BadLocationException;
        
        public abstract String getString(int p0, int p1) throws BadLocationException;
        
        public abstract void getChars(int p0, int p1, Segment p2) throws BadLocationException;
    }
}
